/************************************************
* TOH.C - Tower of Hanoi - an animated solution	*
* to the classic recursion problem.				*
* Martin L. Barrett & Vedula N. Murty			*
************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <graphics.h>
#include <conio.h>
#include <dos.h>

#define xDISKSIZE	 8	/* Horizontal disk size	*/
#define yDISKSIZE	12	/* Vertical disk size	*/
#define	MAXDISKS	10	/* Maximum tower size	*/
#define NUMPEGS		 3	/* Number of pegs		*/
#define FOREGND		 1	/* One disk color		*/
#define FOREGND2	 4	/* Another disk color	*/
#define BACKGND		 0	/* Background color		*/
#define DRAW		 0  /* Draw the disk		*/
#define ERASE		 1	/* Erase the disk		*/

typedef struct {		/* Record for one disk	*/
	int	size,
		color;
 } DISKTYPE;

typedef struct {		/* Record for one peg	*/
	int	NumDisk;
	DISKTYPE disk[MAXDISKS];
 } PEGTYPE;

/* Variables for the graphics display. */
int	xspacing, yspacing, delayrate;

/* Counter for the number of moves. */
int	count = 0;


/************************************************
* drawdisk - draws one disk of peg # pegnum.	*
*												*
* INP:	peg - peg number (0, 1 or 2)			*
*		whichdisk - descr of disk size, color	*
*		disknum - vertical offset height of disk*
*		erase - TRUE to erase, FALSE to draw	*
************************************************/
void drawdisk(int peg, DISKTYPE *whichdisk,
			  int disknum, int erase)
{
int	 upleftx, uplefty, lowrightx, lowrighty;

	/* Compute upper left & lower right corners	*/
	upleftx   = (peg+1)*xspacing -
				whichdisk->size*xDISKSIZE + 2;
	lowrightx = (peg+1)*xspacing +
				whichdisk->size*xDISKSIZE - 2;
	uplefty   = yspacing - (disknum-1)*yDISKSIZE;
	lowrighty = yspacing - disknum*yDISKSIZE + 2;

	/* Draw the disk. */
	setfillstyle(SOLID_FILL,
				erase?BACKGND:whichdisk->color);
	bar(upleftx, uplefty, lowrightx, lowrighty);
	delay(delayrate);	/* Wait before returning*/
}

/************************************************
* movedisk - moves the top disk of peg takeoff	*
* and places it on peg puton.					*
*												*
* INP:	from - peg to move disk from			*
*		to - peg to move disk to				*
*		peg - pointer to array of three pegs	*
************************************************/
void movedisk(int from, int to, PEGTYPE *peg)
{
	/* Erase the disk being moved. */
	drawdisk(from,&peg[from].disk[peg[from].
			NumDisk-1],peg[from].NumDisk-1,ERASE);

	/* Move the disk's record to the new peg */
	peg[to].NumDisk++;
	peg[to].disk[peg[to].NumDisk-1].size =
		peg[from].disk[peg[from].NumDisk-1].size;
	peg[to].disk[peg[to].NumDisk-1].color =
		peg[from].disk[peg[from].NumDisk-1].color;
	peg[from].NumDisk--; /* one less disk on peg */

	/* Draw the disk at its new position. */
	drawdisk(to, &peg[to].disk[peg[to].NumDisk-1],
		 peg[to].NumDisk-1,DRAW);
}

/************************************************
* movetower - recursive to move the tower from	*
* peg 'from' to peg 'to' using peg 'using'.		*
*												*
* INP:	height	- height of tower				*
*		from	- peg to move tower from		*
*		to		- peg to move tower to			*
*		using	- peg to use as 'temporary'		*
*		peg		- ptr to array of pegs			*
* NOTE:	This is the recursive f(n) that solves	*
*		the Tower of Hanoi problem				*
************************************************/
void movetower( int height, int from, int to,
				int using, PEGTYPE *peg)
{
	if (height > 0)
	{
	   /* Use the target peg as intermediate */
	   movetower(height-1,from,using,to, peg);
	   movedisk(from,to, peg);
	   count++;
	   /* Move pegs from intermediate to target */
	   movetower(height-1,using,to,from, peg);
	}
}

/************************************************
* drawalldisks - draws the current position		*
*												*
* INP:	peg - ptr to peg array					*
************************************************/
void drawalldisks(PEGTYPE *peg)
{
int	i, j;

	for (i = 0; i < NUMPEGS; i++)
		for (j = 0; j < peg[i].NumDisk; j++)
			drawdisk(i, &peg[i].disk[j], j, DRAW);
}

/************************************************
*   init_graphics - initializes graphics mode.	*
************************************************/
void init_graphics(void)
{
	int	g_driver, g_mode, g_error;

	g_driver=0; g_mode=0;
	detectgraph(&g_driver, &g_mode);
	if (g_driver < 0) {
		printf("Error: driver not found.\n");
		exit(1);
	}

	/* Initialize graphics mode; quit on error */
	initgraph(&g_driver, &g_mode, "");
	g_error = graphresult();
	if (g_error < 0) {
		printf("Error: initgraph error: %s.\n",
		       grapherrormsg(g_error));
		exit(1);
	}

	yspacing = getmaxy()/2;	/* Set the spacing	*/
	xspacing = getmaxx()/(NUMPEGS+1);
	cleardevice();			/* Clear the screen	*/
}

/************************************************
*	main - do all the work 						*
************************************************/
void main(int argc, char *argv[])
{
	int i, total;
	PEGTYPE	peg[NUMPEGS];

	/* Set delay from command line if given */
	if (argc == 1) delayrate = 100;
	else delayrate = atoi(argv[1]);

	/* Get the number of disks. */
	printf("How many disks? ");
	scanf("%d", &total);
	printf("\n");
	if (total > MAXDISKS) {
	   total = MAXDISKS;
	   printf("Too many disks - using %d\n",total);
	 }

	/* Initialize first peg with all the disks */
	peg[0].NumDisk = total;
	peg[1].NumDisk = peg[2].NumDisk = 0;
	for (i = 0; i < total; i++) {
	   peg[0].disk[i].size = total-i;
	   peg[0].disk[i].color =
	   			(i % 2 == 0) ? FOREGND : FOREGND2;
	}

	/* Initialize graphics & draw starting pos. */
	init_graphics();
	drawalldisks(peg);

	/* Call the recursive procedure. */
	movetower(total, 0, 2, 1, peg);

	/* Wait until a key is pressed, then quit. */
	getch();
	closegraph();
	printf("\n\nThe number of moves was %d\n", count);
}
